import { Token, TokenType } from "./tokenizer.ts";

export interface ASTNode {
    type: string;
}

// export interface VariableDeclarationNode extends ASTNode {
//     type: "VariableDeclaration";
//     identifier: string;
//     value: ASTNode;
//     vtype: string;
//     length: number;
// }

// export interface FunctionDeclarationNode extends ASTNode {
//     type: "FunctionDeclaration";
//     name: string;
//     // params: string[];
//     body: ASTNode[];
// }

export interface AssignmentNode extends ASTNode {
    type: "Assignment";
    identifier: IdentifierNode;
    value: ASTNode;
}

// export interface BinaryExpressionNode extends ASTNode {
//     type: "BinaryExpression";
//     operator: string;
//     left: ASTNode;
//     right: ASTNode;
// }

export interface LiteralNode extends ASTNode {
    type: "Literal";
    value: string;
}

export interface NumberNode extends ASTNode {
    type: "Number";
    value: number;
}

export interface IdentifierNode extends ASTNode {
    type: "Identifier";
    name: string;
}

export interface ExpressionNode extends ASTNode {
    type: "Expression";
    expression: string;
}

export interface LambdaNode extends ASTNode {
    type: "Lambda";
    definition: ExpressionNode | LambdaNode | NumberNode;
    arg: LiteralNode;
}

export interface CallNode extends ASTNode {
    type: "Call";
    function: LambdaNode | CallNode;
    arg: ExpressionNode | NumberNode | CallNode;
}

// export interface BranchFunctionCallNode extends ASTNode {
//     type: "BranchFunctionCall";
//     identifier: string;
//     args: ASTNode[];
//     branches: ASTNode[][];
// }

// export interface StartBlockNode extends ASTNode {
//     type: "StartBlock";
//     body: ASTNode[];
// }

// export interface IfNode extends ASTNode {
//     type: "If";
//     condition: ASTNode;
//     thenBranch: ASTNode[];
//     elseBranch?: ASTNode[];
// }

// export interface WhileNode extends ASTNode {
//     type: "While";
//     condition: ASTNode;
//     branch: ASTNode[];
// }

export interface EvalNode extends ASTNode {
    type: "Eval";
    call: CallNode;
}

export default class AST {
    private tokens: Token[];
    position: number = 0;

    constructor(tokens: Token[]) {
        this.tokens = tokens;
    }

    private peek(ahead = 0): Token {
        return this.tokens[this.position + ahead];
    }

    private advance(): Token {
        return this.tokens[this.position++];
    }

    private match(...types: TokenType[]): boolean {
        if (types.includes(this.peek().type)) {
            console.log(this.peek(), 'matches', types)
            this.advance();
            return true;
        }
        return false;
    }

    private matchTk(types: TokenType[], token = this.peek()): boolean {
        if (types.includes(token.type)) {
            return true;
        }
        return false;
    }

    private expect(type: TokenType, errorMessage: string): Token {
        if (this.peek().type === type) {
            return this.advance();
        }
        console.error('trace: tokens', this.tokens, '\nIDX:', this.position);
        throw new Error(errorMessage);
    }

    parse(): ASTNode[] {
        const nodes: ASTNode[] = [];
        while (this.peek().type !== TokenType.EOF) {
            nodes.push(this.parseStatement());
        }
        return nodes;
    }
    
    private parseStatement(onlyLambda = false): ASTNode {
        console.debug(this.peek())
        // eval
        if (this.match(TokenType.EVAL) && !onlyLambda) {
            const call = this.parseStatement() as CallNode;
            if (call.type != 'Call') throw 'expected call after eval';
            return {
                call,
                type: 'Eval'
            } as EvalNode
        }
        // lambda
        if (this.match(TokenType.LPAREN)) {
            const arg = {
                type: 'Literal',
                value: this.expect(TokenType.IDENTIFIER, 'expected argument in lambda function').value
            } as LiteralNode;
            this.expect(TokenType.DOT, 'expected separator (.) after argument in lambda expression');
            const def: ASTNode = this.parseStatement(false);
            if ((def as ExpressionNode).type != 'Expression' &&
                (def as LambdaNode).type != 'Lambda' &&
                (def as NumberNode).type != 'Number') throw 'wrong definition type';
            const definition: ExpressionNode | LambdaNode | NumberNode =
                def as ExpressionNode | LambdaNode | NumberNode;
            this.expect(TokenType.RPAREN, 'expected end of lambda')
            return {
                type: 'Lambda',
                definition,
                arg,
            } as LambdaNode
        }
        // call
        if (this.match(TokenType.LBRACE) && !onlyLambda) {
            console.debug('uh', this.position, this.peek())
            const lambda = this.parseStatement() as LambdaNode | CallNode;
            if ((lambda as LambdaNode).type != 'Lambda' &&
                (lambda as CallNode).type != 'Call') throw 'wrong lambda type';
            // if (lambda.type != 'Lambda') throw 'expected lambda in call';
            this.expect(TokenType.COMMA, 'expected comma in call');
            const arg = this.parseStatement();
            if ((arg as ExpressionNode).type != 'Expression' &&
                (arg as LambdaNode).type != 'Lambda' &&
                (arg as NumberNode).type != 'Number' &&
                (arg as CallNode).type != 'Call') throw 'wrong arg type';
            this.expect(TokenType.RBRACE, 'expected closing of call')
            return {
                type: 'Call',
                function: lambda,
                arg,
            } as CallNode
        }
        if (this.matchTk([TokenType.LITERAL]) && !onlyLambda) {
            return {
                expression: this.advance().value,
                type: 'Expression'
            } as ExpressionNode
        }
        if (this.matchTk([TokenType.NUMBER]) && !onlyLambda) {
            const value = +this.advance().value;
            if (Number.isNaN(value)) throw 'whar,,, how the fuck is it NaN'
            return {
                value,
                type: 'Number'
            } as NumberNode
        }
        throw `uh oh${onlyLambda ? ' (OL)' : ''}`
    }
}